;; Generated package description from oauth2.el  -*- no-byte-compile: t -*-
(define-package "oauth2" "0.17.0.20250810.115209" "OAuth 2.0 Authorization Protocol" '((emacs "27.1")) :commit "fc1a4b7ac6ae7c2b3596004cef092d767bbf8f8c" :authors '(("Julien Danjou" . "julien@danjou.info")) :maintainer '(nil . "emacs-devel@gnu.org") :keywords '("comm") :url "https://elpa.gnu.org/packages/oauth2.html")
